-----------------------------------------------------------------------------
-- |
-- Module      :  Network.FastCGI
-- Copyright   :  (c) Bjorn Bringert 2004-2005
-- License     :  BSD-style (see the file libraries/network/LICENSE)
-- 
-- Maintainer  :  bjorn@bringert.net
-- Stability   :  experimental
-- Portability :  non-portable (uses FFI)
--
-- Interface for FastCGI <http://fastcgi.com/>, using the fcgiapp API.
--
-- Not MT-safe.
--
-----------------------------------------------------------------------------
module Network.FastCGI ({-runFastCGIConcurrent,-}runFastCGI, runOneFastCGI,
                        module Network.NewCGI) where

import Control.Concurrent (forkOS)
import Control.Exception as Exception (catch, finally, throwIO)
import Control.Monad (when, liftM)
import Data.Char (chr)
import Foreign              ( Ptr, nullPtr, plusPtr, peekArray0, alloca
		            , mallocBytes, free, throwIfNeg_, allocaBytes,
                            , mallocArray, reallocArray)
import Foreign.C            ( CInt, CString,CChar, withCStringLen
                            , peekCString, castCCharToChar )
import Foreign.Storable     ( Storable (..) )
import System.IO.Unsafe (unsafeInterleaveIO)

import Network.NewCGI

import qualified Data.FastPackedString as FPS

-- For debugging
import Control.Concurrent (myThreadId)
import Prelude hiding (log)
import System.IO (hPutStrLn, stderr)

#include <fcgiapp.h>

data FCGX_Stream = FCGX_Stream
type StreamPtr = Ptr FCGX_Stream
type Environ = Ptr CString


foreign import ccall threadsafe "fcgiapp.h FCGX_GetStr" fcgx_getStr
    :: CString -> Int -> StreamPtr -> IO Int

foreign import ccall threadsafe "fcgiapp.h FCGX_PutStr" fcgx_putStr
    :: CString -> Int -> StreamPtr -> IO CInt

--
-- This code uses the not MT safe functions of the fcgiapp API,
-- to avoid getting intermittent SIGPIPEs. Dunno why that happens.
--

foreign import ccall "fcgiapp.h FCGX_Accept" fcgx_accept
    :: Ptr StreamPtr -> Ptr StreamPtr -> Ptr StreamPtr -> Ptr Environ -> IO CInt
foreign import ccall "fcgiapp.h FCGX_Finish" fcgx_finish
    :: IO ()

-- | Handle FastCGI requests in an infinite loop.
runFastCGI :: CGI CGIResult -> IO ()
runFastCGI f = runOneFastCGI f >> runFastCGI f

-- | Use a CGI action to handle a single FastCGI request.
runOneFastCGI :: CGI CGIResult -> IO ()
runOneFastCGI f = do
    alloca (\inp ->
            alloca (\outp ->
                    alloca (\errp ->
                            alloca (\envp ->
                                    oneRequest f inp outp errp envp))))

oneRequest :: CGI CGIResult
           -> Ptr StreamPtr
           -> Ptr StreamPtr
           -> Ptr StreamPtr
           -> Ptr Environ
           -> IO ()
oneRequest f inp outp errp envp =
    do
    testReturn "FCGX_Accept" $ fcgx_accept inp outp errp envp
    ins  <- peek inp
    outs <- peek outp
    errs <- peek errp
    env  <- peek envp
    handleRequest f ins outs errs env
    fcgx_finish

handleRequest :: CGI CGIResult
              -> StreamPtr
              -> StreamPtr
              -> StreamPtr
              -> Environ
              -> IO ()
handleRequest f ins outs errs env =
    do
    vars <- environToTable env
    input <- sRead ins
    output <- runCGIEnvFPS vars input f
    sPutStr outs output



--
-- The following code uses the MT safe fcgiapp API,
-- but causes mod_fastcgi to get SIGPIPE once in a while.
--

{-

data FCGX_Request = FCGX_Request

foreign import ccall threadsafe "fcgiapp.h FCGX_Init" fcgx_init
    :: IO CInt

foreign import ccall threadsafe "fcgiapp.h FCGX_InitRequest" fcgx_initrequest
    :: Ptr FCGX_Request -> CInt -> CInt -> IO CInt

foreign import ccall threadsafe "fcgiapp.h FCGX_Accept_r" fcgx_accept_r
    :: Ptr FCGX_Request -> IO CInt

foreign import ccall threadsafe "fcgiapp.h FCGX_Finish_r" fcgx_finish_r
    :: Ptr FCGX_Request -> IO ()

-- | Like 'Network.NewCGI.runCGI', but uses the FastCGI interface.
runFastCGI :: CGI CGIResult -> IO ()
runFastCGI = runFastCGI' id

-- | Like 'Network.NewCGI.runCGI', but uses the FastCGI interface
--   and forks off a new thread (using 'forkOS') for every request.
runFastCGIConcurrent :: CGI CGIResult -> IO ()
runFastCGIConcurrent = runFastCGI' fork
    where fork m = do
		   t <- forkOS m
		   log $ "Created child: " ++ show t

runFastCGI' :: (IO () -> IO a) -> CGI CGIResult -> IO ()
runFastCGI' fork f = do
		     log "Calling FCGX_Init"
		     testReturn "FCGX_Init" $ fcgx_init
		     loop
    where 
    loop = do
	   reqp <- acceptRequest
	   log "Forking"
	   fork (oneRequest f reqp `finally` finishRequest reqp)
	   loop

oneRequest :: CGI CGIResult -> Ptr FCGX_Request -> IO ()
oneRequest f r = do
		 env <- peekEnvp r
		 vars <- environToTable env
		 ins <- peekIn r
		 input <- sStrictRead ins
		 log "Running CGI action"
		 output <- runCGIEnv vars input f
		 outs <- peekOut r
		 log "Returning output"
		 sPutStr outs output

--
-- * FCGX_Reqest struct
--

acceptRequest :: IO (Ptr FCGX_Request)
acceptRequest = do
		log "Allocating FCGX_Request"
 		reqp <- mallocBytes (#size FCGX_Request)
		initAndAccept reqp `onError` free reqp
		return reqp
    where initAndAccept reqp = 
	      do 
	      log "Calling FCGX_InitRequest"
	      testReturn "FCGX_InitRequest" $ fcgx_initrequest reqp 0 0
	      log "Calling FCGX_Accept_r"
	      testReturn "FCGX_Accept_r" $ fcgx_accept_r reqp

finishRequest :: Ptr FCGX_Request -> IO ()
finishRequest reqp = do
		     log "Calling FCGX_Finish_r"
		     fcgx_finish_r reqp
		     log "Freeing FCGX_Request"
		     free reqp

peekIn, peekOut, peekErr :: Ptr FCGX_Request -> IO (Ptr FCGX_Stream)
peekIn  = (#peek FCGX_Request, in)
peekOut = (#peek FCGX_Request, out)
peekErr = (#peek FCGX_Request, err)

peekEnvp :: Ptr FCGX_Request -> IO Environ
peekEnvp = (#peek FCGX_Request, envp)

-}

--
-- * Stream IO
--

-- Based on Data.FastPackedStirng.hGetContents
sRead :: StreamPtr -> IO FPS.FastString
sRead h = do 
    let start_size = 1024
    p <- mallocArray start_size
    i <- fcgx_getStr p start_size h
    if i < start_size
        then liftM FPS.packMallocCString $ reallocArray p i
        else f p start_size
    where 
        f p s = do 
        let s' = 2 * s
        p' <- reallocArray p s'
        i  <- fcgx_getStr (p' `plusPtr` s) s h
        if i < s 
            then liftM FPS.packMallocCString $ reallocArray p' (s + i)
            else f p' s'

sPutStr :: StreamPtr -> String -> IO ()
sPutStr out str = 
    withCStringLen str (\(cs,len) -> testReturn "FCGX_PutStr" $ fcgx_putStr cs len out)


--
-- * Utilities
--

-- | Run some action if the first action throws an exception.
--   The exception is re-thrown.
onError :: IO a -> IO () -> IO a
onError f h = f `Exception.catch` (\e -> h >> throwIO e)

testReturn :: String -> IO CInt -> IO ()
testReturn e = throwIfNeg_ (const $ e ++ " failed")

environToTable :: Environ -> IO [(String,String)]
environToTable arr = 
    do css <- peekArray0 nullPtr arr
       ss <- mapM peekCString css
       return $ map (splitBy '=') ss

-- | Split a list at the first occurence of a marker.
--   Do not include the marker in any of the resulting lists.
--   If the marker does not occur in the list, the entire
--   input with be in the first list.
splitBy :: Eq a => a -> [a] -> ([a],[a])
splitBy x xs = (y, drop 1 z)
    where (y,z) = break (==x) xs

--
-- * Debugging
--

log :: String -> IO ()
log msg = do
	  t <- myThreadId
	  hPutStrLn stderr (show t ++ ": " ++ msg)